import os
import socket

HSMM_PORT = 5000
LEN_BUFF = 512


class HSMMConnectionException(Exception):
    pass


class HSMMConnection:
    def __init__(
        self,
        hostname: str | None = None,
    ) -> None:
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if hostname is None:
            hostname = os.environ.get("HSMM_HOSTNAME", None)
            if hostname is None:
                raise HSMMConnectionException(
                    "Environment variable HSMM_HOSTNAME is required"
                )
        try:
            self.sock.connect((hostname, HSMM_PORT))
        except ConnectionRefusedError:
            self.sock.close()
            raise HSMMConnectionException("Cannot connect to HSMM")

    def close(self) -> None:
        self.sock.close()

    def sendline(
        self,
        msg: str,
    ) -> None:
        msgline = f"{msg}\n".encode()
        len_msgline = len(msgline)
        totalsent = 0
        while totalsent < len_msgline:
            sent = self.sock.send(msgline[totalsent:])
            if sent == 0:
                raise HSMMConnectionException("HSMMConnection broken")
            totalsent = totalsent + sent

    def sendline(
        self,
        msg: str | bytes,
    ) -> None:
        if type(msg) == str:
            msgline = f"{msg}\n".encode()
        elif type(msg) == bytes:
            msgline = msg + "\n".encode()
        else:
            raise HSMMConnectionException("sendline() requires str or bytes msg")
        len_msgline = len(msgline)
        totalsent = 0
        while totalsent < len_msgline:
            sent = self.sock.send(msgline[totalsent:])
            if sent == 0:
                raise HSMMConnectionException("HSMMConnection broken")
            totalsent = totalsent + sent

    def recvlines(
        self,
    ) -> list[str]:
        msgline = ""
        while True:
            chunk = self.sock.recv(LEN_BUFF)
            if chunk == b"":
                raise HSMMConnectionException("HSMMConnection broken")
            msgline += chunk.decode()
            if msgline[-1] == "\n":
                break
        return msgline[:-1].split("\n")


class HSMMException(Exception):
    def __init__(self, message=None):
        if (message is not None) and (not "HSM Error:" in message):
            message = "Unknown HSM Error"
        super().__init__(message)


def sign(
    user: str,
    pwd: str,
    recipient: str,
    content: bytes,
    extra_0: int,
    extra_1: int,
) -> tuple[str, str]:
    conn = HSMMConnection()
    try:
        conn.sendline("0")
        conn.sendline(user)
        conn.sendline(pwd)
        conn.sendline(recipient)
        conn.sendline(content)
        conn.sendline(f"{extra_0}")
        conn.sendline(f"{extra_1}")
        recvs = conn.recvlines()
        if len(recvs) != 2:
            if len(recvs) == 1:
                raise HSMMException(recvs[0])
            raise HSMMException()
        data_b64 = recvs[0]
        sig_b64 = recvs[1]
        return (data_b64, sig_b64)
    finally:
        conn.close()


def verify(
    data: bytes,
    sig_b64: str,
) -> bool:
    conn = HSMMConnection()
    try:
        conn.sendline("1")
        conn.sendline(sig_b64)
        conn.sendline(data)
        recvs = conn.recvlines()
        if len(recvs) != 1:
            HSMMException()
        recv = recvs[0]
        if recv == "0":
            return False
        elif recv == "1":
            return True
        raise HSMMException(recv)
    finally:
        conn.close()
